//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <fstream>
#include "MapGenerator.h"
#include "file.h"
#include "defs.h"

using namespace std;


MapGenerator::MapGenerator()
    //set reasonable default values
    :d_pswamp(2), d_pwater(7), d_pforest(3), d_phills(5), d_pmountains(5),
    d_nocities(20), d_notemples(25), d_noruins(25)
{
	// initialize the building array
	for(int x=0;x<100;x++)
		for(int y=0;y<100;y++)
			d_building[x][y]=' ';
	
	//load file for city names
	ifstream namefile(File::getMiscFile("citynames").c_str());

	int counter;
	namefile >> counter;
	d_namelist.resize(counter);

	//with getline, the first call will return the first line to the end
	//of the line, i.e. a newline character. To prevent this, we do a throw
	//away call first
	char buffer[100];
	namefile.getline(buffer, 100);
	
	for(counter--; counter >= 0; counter--)
	{
		namefile.getline(buffer, 100);

		d_namelist[counter] = std::string(buffer);
	}
}

MapGenerator::~MapGenerator()
{
}

int MapGenerator::setNoCities(int nocities)
{
    if (nocities <= 0)
        return -1;

    int tmp = d_nocities;
    d_nocities = nocities;
    return tmp;
}

int MapGenerator::setNoRuins(int noruins)
{
    if (noruins < 0)
        return -1;

    int tmp = d_noruins;
    d_noruins = noruins;
    return tmp;
}

int MapGenerator::setNoTemples(int notemples)
{
    if (notemples < 0)
        return -1;

    int tmp = d_notemples;
    d_notemples = notemples;
    return tmp;
}

void MapGenerator::setPercentages(int pwater, int pforest, int pswamp,
                                    int phills, int pmountains)
{
    if ((pswamp < 0) || (pwater < 0) || (pforest < 0) || (phills < 0)
        || (pmountains < 0))
        return;

    if (pswamp + pwater + pforest + phills + pmountains > 100)
        return;

    d_pswamp = pswamp;
    d_pwater = pwater;
    d_pforest = pforest;
    d_phills = phills;
    d_pmountains = pmountains;
}

/** 
 * Generates a random map . The map is stored as a char array of size 
 * 100x100. Each character stands for something. We use :
 * M = mountains
 * h = hills
 * ~ = water
 * $ = forest
 * . = plains
 * _ = swamps
 * C = city/castle
 * r = ruins
 * T = temple
 *   = nothing
 * c = part of city/castle
 */
void MapGenerator::MakeMap(int xsize,int ysize)
{
    d_xsize = xsize;
    d_ysize = ysize;

    cout<<"Making random map"<<endl;
   
    cout<<"Flatening Plains  ... 10%"<<endl;
    makePlains();
    cout<<"Raising Hills     ... 20%"<<endl;
    makeTerrain(CHARACTER_HILLS,d_phills);
    cout<<"Raining Water     ... 30%"<<endl;
    makeTerrain(CHARACTER_WATER,d_pwater);
    cout<<"Planting Forest   ... 40%"<<endl;
    makeTerrain(CHARACTER_FOREST,d_pforest);
    cout<<"Watering Swamps   ... 50%"<<endl;
    makeTerrain(CHARACTER_SWAMP,d_pswamp);
    cout<<"Creating Mountains ... 60%"<<endl;
    makeTerrain(CHARACTER_MOUNTAINS,d_pmountains);
    
    cout<<"Building Cities   ... 70%"<<endl;
    makeCities(d_nocities);
    cout<<"Ruining Ruins     ... 80%"<<endl;
    makeBuildings(CHARACTER_RUIN,d_noruins);
    cout<<"Spawning temples  ... 90%"<<endl;
    makeBuildings(CHARACTER_TEMPLE,d_notemples);
    cout<<"Done making map   ... 100%"<<endl;
}

void MapGenerator::DumpMap(const char* filename)
{
    ofstream ofile(filename);
    int i,j;
    cout<<"Dumping map "<<filename<<endl;
    for(i =0;i<d_xsize;i++)
    {
        for(j=0;j<d_ysize;j++)
        {
            ofile<<d_terrain[i][j];
        }
        ofile<<endl;
    }
    for(i =0;i<d_xsize;i++)
    {
        for( j=0;j<d_ysize;j++)
        {
            ofile<<d_building[i][j];
        }
        ofile<<endl;
    }

    //now write down the city names
    for (unsigned int i = 0; i < d_citynames.size(); i++)
    	ofile <<d_citynames[i] <<endl;

    ofile.close();
}

char* MapGenerator::getMap(int& xsize, int& ysize)
{
    xsize = d_xsize;
    ysize = d_ysize;

    return d_terrain[0];
}

char* MapGenerator::getBuildings(int& xsize, int& ysize)
{
    xsize = d_xsize;
    ysize = d_ysize;

    return d_building[0];
}

std::vector<std::string>* MapGenerator::getCitynames()
{
    return &d_citynames;
}

/* Nothing much to do here, just fill up everything with '.' 
 * as we are making the default canvas for the map.
 */
void MapGenerator::makePlains()
{
    for(int i =0;i<d_xsize;i++)
    {
        for(int j=0;j<d_ysize;j++)
        {
            d_terrain[i][j]=CHARACTER_GRASS;
        }
    }
}


/**
 * Makes Terrains.
 * The algo is as follows :
 * 1. Find a random starting location
 * 2. chose a random direction , if x is the starting location, the direction
 *    can be from 0-7 as follows :
 *    +-+-+-+
 *    |0|1|2|
 *    +-+-+-+
 *    |7|x|3|
 *    +-+-+-+
 *    |6|5|4|
 *    +-+-+-+
 * 3. Check if there is some other terrain there ('.' = plains is okay)
 * 4. Reserve the location, and chose another direction
 */
void MapGenerator::makeTerrain(char c,int percent)
{
    // calculate the total number of tiles for this terrain
    int terrain = d_xsize*d_ysize*percent / 100;
    int xdir[]= {-1, 0, 1, 1, 1, 0,-1,-1};
    int ydir[]= { 1, 1, 1, 0,-1,-1,-1, 0};

    int count = 0;  // counter 
    int done = 0;
    int loopcount = 0;
    int tmpx = 0;
    int tmpy = 0;
    int direction = 0;
    int x,y;    // locations
    while(count != terrain)
    {
        // find a random starting location
        do{
            x = rand()%100;
            y = rand()%100;
        }while( d_terrain[x][y]!=CHARACTER_GRASS);
        // starting position
        d_terrain[x][y]=c;
        count++;
        while(count != terrain)
        {
            //find direction
            done = 0,loopcount = 0;
            direction  = rand()%8;
            while( !done )
            {
                direction +=loopcount;
                // NOTE: potential hazard here as if the direction
                // becomes more than the max value, it will always be
                // set to 0 (north west).
                if(direction>7)
                    direction = 0;
                tmpx = x+xdir[direction];
                tmpy = y+ydir[direction];
                if(d_terrain[tmpx][tmpy]==CHARACTER_GRASS)
                    done =1;
                if(++loopcount > 7)
                    done =1;
                

            }
            if(loopcount>=7 || tmpx <0 || tmpy <0 || tmpx > 100 || tmpy >100)
                break; // find another starting point
            x=tmpx;
            y=tmpy;
            d_terrain[x][y]=c;
            count++;
        }

    }// end while
    
    
}
void MapGenerator::makeCities(int cities)
{
    int x,y,city_count;
    
    // place the cities
    city_count = 0;
    while(city_count < cities)
    {
        x = rand()%(d_xsize-2);
        y = rand()%(d_ysize-2);
        if(d_building[x][y] != ' ')
            continue;

        // check if we can put the building
        if(!canPutBuilding(x,y))
            continue;
        
        d_building[x][y]    =CHARACTER_CITY_MAJOR;
        d_building[x+1][y]  =CHARACTER_CITY_MINOR;
        d_building[x][y+1]  =CHARACTER_CITY_MINOR;
        d_building[x+1][y+1]=CHARACTER_CITY_MINOR;
        city_count ++;
    }

    //give the cities names (order from top left to down right)
    d_citynames.resize(cities);

    if (d_citynames.size() > d_namelist.size())
    //mustn't happen that there are more cities than names for them
    {
    	std::cerr<<"Not enough names for naming cities on random map, exiting\n";
        std::cerr<<d_namelist.size() <<" names existed for " <<city_count <<" cities\n";
        exit(-1);
    }

    for (city_count = 0; city_count < cities; city_count++)
    {
    	int rand_no;
        rand_no = rand() % d_namelist.size();
        d_citynames[city_count] = d_namelist[rand_no];
        d_namelist[rand_no] = d_namelist[d_namelist.size()-1];
        d_namelist.pop_back();	//delete the already used name
    }
}

void MapGenerator::makeBuildings(char c, int building)
{
    int x,y;
    
    for(int i=0;i<building;i++)
    {
        do{
            x = rand()% d_xsize;
            y = rand()% d_ysize;
        }while( d_building[x][y]!=' ' || d_terrain[x][y]==CHARACTER_WATER);
        d_building[x][y]=c;
    }
}


/** 
 * canPutBuilding
 * Checks if we can put a building at the specified place. 
 * If we are on a sqare with water, we cannot put it.
 * If out of the surrounding squares, we have water in more than
 * 8 squares, we cannot.
 * int x = x position
 * int y = y position
 */

bool MapGenerator::canPutBuilding(int x,int y)
{
    int count=0;        // stores the count of water squares.
    // if the building is on water, return false
    if(d_terrain[x][y] == CHARACTER_WATER || d_terrain[x+1][y] == CHARACTER_WATER 
        || d_terrain[x][y+1] == CHARACTER_WATER || d_terrain[x+1][y+1] == CHARACTER_WATER)
        return false;
    x = x-1;y=y-1;
    if(x<0)     x = 0;
    if(x>96)    x = 96;
    if(y<0)     y = 0;
    if(y>96)    y = 96;
        
    // check for water in surrounding sqares. 
    for(int i=x;i<x+4;i++)
        for(int j=y;j<y+4;j++)
            if(d_terrain[x][y] == CHARACTER_WATER)
                count++;
    if(count>8)
        return false;
    // everything okay here ! return true
    return true;
    
}
