//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "bigmap.h"
#include <SDL_image.h>
#include <assert.h>
#include "cityinfo.h"
#include "army.h"
#include "armyset.h"
#include "path.h"
#include "mapdef.h"
#include "stacklist.h"
#include "citylist.h"
#include "ruin.h"
#include "temple.h"
#include "playerlist.h"
#include "defs.h"
#include "file.h"

using namespace std;

//#define debug(x) {cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl;}
#define debug(x)

BigMap::BigMap(PG_Widget* parent, PG_Rect rect)
	:PG_Widget(parent, rect, true), d_viewrect(0)
{
	d_enable = true;

	// load all pictures
    d_arrows = File::getMiscPicture("arrows.png");
    d_ruinpic = File::getMapsetPicture("default", "ruin.png");
	d_templepic = File::getMapsetPicture("default", "temple.png");
	d_swordpic = File::getMiscPicture("sword.png");
	d_selector[0] = File::getMiscPicture("select0000.png");
	d_selector[1] = File::getMiscPicture("select0001.png");
	d_polepic = File::getMiscPicture("flag_pole.png");

    char str[80];
    for (int i = 0; i < 9; i++)
    {
        if (i < 8)
        {
            sprintf(str, "small_castle_%i.png", i);
            d_citypic[i] = File::getMapsetPicture("default", str);
        }

        sprintf(str, "flag_%i.png", i);
        SDL_Surface* flag = File::getMiscPicture(str);
        SDL_PixelFormat* fmt = flag->format;
        // !!! very important --- copy alpha values, don't use them
        SDL_SetAlpha(flag, 0, 0);
        for (int j = 0; j < 8; j++)
        {
            SDL_Surface* tmp = SDL_CreateRGBSurface(SDL_HWSURFACE,
                    50, 7, fmt->BitsPerPixel, fmt->Rmask, fmt->Gmask,
                    fmt->Bmask, fmt->Amask);

            SDL_BlitSurface(flag, PG_Rect(0, 7*j, 50, 7).SDLRect(), tmp, 0);
            SDL_Surface* pixmap = SDL_DisplayFormatAlpha(tmp);
            SDL_FreeSurface(tmp);

            d_flagpic[i][j] = pixmap;
        }
        SDL_FreeSurface(flag);
    }

	d_citypic[8] = File::getMapsetPicture("default", "small_castle_neutral.png");
    d_renderer = new MapRenderer(GetWidgetSurface(), mapdef);

    d_timerID = AddTimer(500);
}

BigMap::~BigMap()
{
    RemoveTimer(d_timerID);
    SDL_FreeSurface(d_ruinpic);
    SDL_FreeSurface(d_templepic);
    SDL_FreeSurface(d_swordpic);
    SDL_FreeSurface(d_selector[0]);
    SDL_FreeSurface(d_selector[1]);
    SDL_FreeSurface(d_polepic);

	for (int i = 0; i < 9; i++)
	{
	    SDL_FreeSurface(d_citypic[i]);
		for (int j = 0; j < 8; j++)
		{
            SDL_FreeSurface(d_flagpic[i][j]);
		}
	}
    delete d_renderer;
}

void BigMap::eventDraw(SDL_Surface* surface, const PG_Rect& rect)
{
	debug("eventDraw()");

    // Draw map
    d_renderer->render(0, 0, d_viewrect->my_xpos, d_viewrect->my_ypos,
            d_viewrect->my_width, d_viewrect->my_height);


    // Draw all ruins
    for (Ruinlist::iterator it = ruinlist->begin(); it != ruinlist->end(); ++it)
    {
        PG_Point p = (*it).getPos();
        if (d_viewrect->IsInside(p))
        {
            p = getRelativePos(p);
            SDL_BlitSurface(d_ruinpic, 0, surface, PG_Rect(p.x, p.y, 50, 50).SDLRect());
        }
    }

    // Draw temples
    for (TempleList::iterator it = templeList->begin(); it != templeList->end(); ++it)
    {
        PG_Point p = (*it).getPos();
        if (d_viewrect->IsInside(p))
        {
            p = getRelativePos(p);
            SDL_BlitSurface(d_templepic, 0, surface, PG_Rect(p.x, p.y, 50, 50).SDLRect()); 
        }
    }

    // Draw all cities
    for (Citylist::iterator it = citylist->begin(); it != citylist->end(); ++it)
    {	
        PG_Point p = (*it).getPos();
        if (d_viewrect->IsInside(p))
        {
            p = getRelativePos(p);

            // set picture for normal or neutral player
            int nr = playerlist->getPlayerNr((*it).getPlayer());
            if (nr == -1) nr = 8;
            SDL_BlitSurface(d_citypic[nr], 0, surface, PG_Rect(p.x, p.y, 100, 100).SDLRect());
        }
    }

    // Draw stacks
    for (Stacklist::iterator it = stacklist->begin(); it != stacklist->end(); it++)
    {
        PG_Point p = (*it)->getPos();

        // check if the object lies in the viewed part of the map
        // otherwise we shouldn't draw it
        if (d_viewrect->IsInside(p))
        {
            p = getRelativePos(p);

            // draw stack
            SDL_BlitSurface((*it)->getStrongestArmy()->getPixmap(), 0,
                    surface, PG_Rect(p.x + 6, p.y + 6, 40, 40).SDLRect());

            // draw flagpole
            SDL_BlitSurface(d_polepic, 0, surface, PG_Rect(p.x, p.y, 50, 50).SDLRect());

            // draw flag
            int nr = playerlist->getPlayerNr((*it)->getPlayer());
            if (nr == -1) nr = 8;	// for neutral player

            if (d_flagpic[nr][(*it)->size()-1]) 
            {
                SDL_BlitSurface(d_flagpic[nr][(*it)->size()-1], 0, surface,
                        PG_Rect(p.x, p.y, 50, 50).SDLRect());
            }
        }
    }

    DrawBorder(rect, 1);

    Stack* stack = stacklist->getActivestack();
    if (!stack) return;

    // Draw Path
    if (stack->getPath()->size())
    {
        PG_Point pos;
        PG_Point nextpos;
        SDL_Color c;
        c.r = c.g = c.b = 0;

        // draw all waypoints except the last
        for (list<PG_Point*>::iterator it = stack->getPath()->begin();
                it != --(stack->getPath()->end());)
        {
            // peak at the next waypoint to draw the correct arrow
            pos = getRelativePos(**it);
            nextpos = getRelativePos(**(++it));
            if (nextpos.x > pos.x && nextpos.y == pos.y)
            {
                SDL_BlitSurface(d_arrows, PG_Rect(100, 0, 50, 50).SDLRect(),
                        surface, PG_Rect(pos.x, pos.y, 50, 50).SDLRect());
            }
            else if (nextpos.x < pos.x && nextpos.y == pos.y)
            {
                SDL_BlitSurface(d_arrows, PG_Rect(300, 0, 50, 50).SDLRect(),
                        surface, PG_Rect(pos.x, pos.y, 50, 50).SDLRect());
            }
            else if (nextpos.x < pos.x && nextpos.y > pos.y)
            {
                SDL_BlitSurface(d_arrows, PG_Rect(250, 0, 50, 50).SDLRect(),
                        surface, PG_Rect(pos.x, pos.y, 50, 50).SDLRect());
            }
            else if (nextpos.x < pos.x && nextpos.y < pos.y)
            {
                SDL_BlitSurface(d_arrows, PG_Rect(350, 0, 50, 50).SDLRect(),
                        surface, PG_Rect(pos.x, pos.y, 50, 50).SDLRect());
            }
            else if (nextpos.x > pos.x && nextpos.y < pos.y)
            {
                SDL_BlitSurface(d_arrows, PG_Rect(50, 0, 50, 50).SDLRect(),
                        surface, PG_Rect(pos.x, pos.y, 50, 50).SDLRect());
            }
            else if (nextpos.x > pos.x && nextpos.y > pos.y)
            {
                SDL_BlitSurface(d_arrows, PG_Rect(150, 0, 50, 50).SDLRect(),
                        surface, PG_Rect(pos.x, pos.y, 50, 50).SDLRect());
            }
            else if (nextpos.x == pos.x && nextpos.y > pos.y)
            {
                SDL_BlitSurface(d_arrows, PG_Rect(200, 0, 50, 50).SDLRect(),
                        surface, PG_Rect(pos.x, pos.y, 50, 50).SDLRect());
            }
            else if (nextpos.x == pos.x && nextpos.y < pos.y)
            {
                SDL_BlitSurface(d_arrows, PG_Rect(0, 0, 50, 50).SDLRect(),
                        surface, PG_Rect(pos.x, pos.y, 50, 50).SDLRect());
            }
        }

        pos = getRelativePos(*stack->getPath()->back());
        SDL_BlitSurface(d_arrows, PG_Rect(400, 0, 50, 50).SDLRect(),
                surface, PG_Rect(pos.x, pos.y, 50, 50).SDLRect());
    }
    drawSelector(false);
}

bool BigMap::eventMouseButtonDown(const SDL_MouseButtonEvent* event)
{

    //pretend to handle event, but don't do it (right now we're
    //waiting for the AI or network guy)
	if (!d_enable)
    {
        return true;
    }

	int pos_X = (event->x - my_xpos) / 50 + d_viewrect->my_xpos;
	int pos_Y = (event->y - my_ypos) / 50 + d_viewrect->my_ypos;

	if (event->button == SDL_BUTTON_LEFT)
	{
		Stack* stack = stacklist->getActivestack();

		if (stack)
		{
			// clicked on an already active stack
			if (stack->getPos().x == pos_X && stack->getPos().y == pos_Y) return true;

			PG_Point p;
			p.x = pos_X; p.y = pos_Y;

			playerlist->getActiveplayer()->stackSplit(stack); // split if ungrouped
			stack->getPath()->calculate(stack, p);
			Redraw();
			playerlist->getActiveplayer()->stackMove(stack);
		}
		// Stack hasn't been active yet
		else
		{
			stack = stacklist->getObjectAt(pos_X, pos_Y);
			if ((stack) && (stack->isFriend(playerlist->getActiveplayer()))) 
			{
				stacklist->setActivestack(stack);
				stackSelected();
			}

		}
        return true;
	}

	// right mousebutton to click at city
	if (event->button == SDL_BUTTON_RIGHT)
	{
		City* city = citylist->getObjectAt(pos_X, pos_Y);

		if ((city) && (city->isFriend(playerlist->getActiveplayer())))
		{
			CityInfo cityinfo(city);
			cityinfo.Show();
			cityinfo.RunModal();
			cityinfo.Hide();
		}
        return true;
	}

	return false;
}

void BigMap::centerView(const PG_Point& p)
{
	debug("centerView()");

	d_viewrect->my_xpos = p.x - (int)(d_viewrect->my_width / 2);
	d_viewrect->my_ypos = p.y - (int)(d_viewrect->my_height / 2);

	if (d_viewrect->my_xpos < 0) d_viewrect->my_xpos = 0;
	if (d_viewrect->my_ypos < 0) d_viewrect->my_ypos = 0;
	if (d_viewrect->my_xpos > (100 - d_viewrect->my_width)) 
        d_viewrect->my_xpos = 100 - d_viewrect->my_width;
	if (d_viewrect->my_ypos > (100 - d_viewrect->my_height))
        d_viewrect->my_ypos = 100 - d_viewrect->my_height;

    schangingViewrect.emit();
}

void BigMap::stackMoved()
{
	debug("stackMoved()");

	Stack* stack = stacklist->getActivestack();
	assert(stack != 0);

    // center the view around the new stack position
	centerView(stack->getPos());
	Redraw();
}

void BigMap::stackSelected()
{
	debug("stackSelected()");

	Stack* stack = stacklist->getActivestack();
	if (!stack->getPath()->checkPath(stack))
	{
		//error handling is required here, up to now we only barf on cerr
		cerr << "original path of stack was blocked\n";
	}

	centerView(stack->getPos());
	Redraw();

    sselectingStack.emit(stack);
}

void BigMap::stackDeselected()
{
	debug("stackDeselected()");
	stacklist->setActivestack(0);
	Redraw();
}

// calculate the relative pos from the absolute pos by subtracting
// the viewrect-pos and multiplying with MAPTILE_SIZE (50)

const PG_Point BigMap::getRelativePos(const PG_Point& absolutePos)
{
	PG_Point relativePos;
	relativePos.x = (absolutePos.x - d_viewrect->my_xpos) * 50;
	relativePos.y = (absolutePos.y - d_viewrect->my_ypos) * 50;
	return relativePos;
}

Uint32 BigMap::eventTimer(PG_TimerID id, Uint32 interval)
{
    drawSelector(true);
    PG_TimerObject::eventTimer(id, interval);
	return interval;
}

// Draw Selector
void BigMap::drawSelector(bool changeFrame)
{
    debug("drawSelector()");
    Stack* stack = stacklist->getActivestack();
    if (!stack) return;
	PG_Point p = stack->getPos();
	if (d_viewrect->IsInside(p))
	{
        static int frame = 0;
        if (frame == 0) frame = 1;
        else frame = 0;
		p = getRelativePos(p);
		SDL_BlitSurface(d_selector[frame], 0, GetWidgetSurface(), PG_Rect(p.x, p.y, 50, 50).SDLRect());
        SDL_UpdateRect(GetWidgetSurface(), p.x, p.y, 50, 50);
        Update();
	}
}

// End of file
