//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <stdlib.h>
#include <algorithm>
#include <assert.h>
#include <fstream>
#include <sstream>
#include "player.h"
#include "playerlist.h"
#include "armysetlist.h"
#include "FightDialog.h"
#include "stacklist.h"
#include "citylist.h"
#include "path.h"
#include "mapdef.h"
#include "armysetlist.h"
#include "real_player.h"
#include "ai_dummy.h"
#include "ai_fast.h"
#include "counter.h"

using namespace std;

#define debug(x) {cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl;}
//#define debug(x)

Player::Player(string name, string armyset, SDL_Color color, TYPE type)
	:d_color(color), d_name(name), d_armyset(armyset), d_gold(1000),
	d_dead(false), d_type(type)
{
    d_id= fl_counter->getNextId();
}

Player::Player(const Player& player)
	:d_color(player.d_color), d_name(player.d_name), d_armyset(player.d_armyset),
	d_gold(player.d_gold), d_dead(player.d_dead), d_type(player.d_type),
	d_actions(player.d_actions)
{
}

Player::Player(XML_Helper* helper)
{
	helper->getData(d_id, "id");
	helper->getData(d_name, "name");
	helper->getData(d_gold, "gold");
	helper->getData(d_dead, "dead");
	
	int i;
	helper->getData(i, "type");
	d_type = static_cast<TYPE>(i);

	string s;
	helper->getData(s, "color");
	
	istringstream scolor(s);
	scolor >>d_color.r >>d_color.g >>d_color.b;

	helper->getData(d_armyset, "armyset");
	armysetlist->loadArmyset(d_armyset);

	//last but not least, register the load function for actionlist
	helper->registerTag("action", SigC::slot((*this), &Player::load));
}

Player::~Player()
{
}

void Player::addGold(int gold)
{
	d_gold += gold;
    schangingStatus.emit();
}

void Player::withdrawGold(int gold)
{
	d_gold -= gold;
    schangingStatus.emit();
}

Armyset* Player::getArmyset()
{
	Armyset* armyset = armysetlist->getArmyset(d_armyset);
	assert(armyset != 0);
	return armyset;
}

void Player::dumpActionlist()
{
	for (list<Action*>::iterator it = d_actions.begin();
		it != d_actions.end(); it++)
	{
		cerr <<(*it)->dump() << endl;
	}	
}

void Player::clearActionlist()
{
	for (list<Action*>::iterator it = d_actions.begin();
		it != d_actions.end(); it++)
    {
		delete (*it);
    }
	d_actions.clear();
}

bool Player::enoughMoves(Stack* s)
{
	PG_Point dest = *(s->getPath()->front());
	
    // A city on the next square
	if (citylist->getObjectAt(dest.x, dest.y))
    {
        return (s->getGroupMoves() >= 1);
    }

	unsigned int maptype = mapdef->map[dest.x][dest.y]->getMaptileType();
	int needed_moves = mapdef->map[dest.x][dest.y]->getMoves();
	int cur_moves;

	for (Stack::iterator it = s->begin(); it != s->end(); it++)
	{
		cur_moves = needed_moves;
		if ((*it)->getMoveBonus() == maptype)
			cur_moves = 2;

		if ((*it)->get_moves() < cur_moves)
			return false;
	}

    return true;
}

bool Player::save(XML_Helper* helper)
{
	bool retval = true;

	stringstream s;
	s <<static_cast<unsigned int>(d_color.r);
	s <<" " <<static_cast<unsigned int>(d_color.g);
	s <<" " <<static_cast<unsigned int>(d_color.b);

	retval &= helper->openTag("player");
	retval &= helper->saveData("id", d_id);
	retval &= helper->saveData("name", d_name);
	retval &= helper->saveData("color", s.str());
	retval &= helper->saveData("armyset", d_armyset);
	retval &= helper->saveData("gold", d_gold);
	retval &= helper->saveData("dead", d_dead);
	retval &= helper->saveData("type", d_type);

	//save the actionlist
	for (list<Action*>::iterator it = d_actions.begin();
			it != d_actions.end(); it++)
		retval &= (*it)->save(helper);

	retval &= helper->closeTag();

	return retval;
}

Player* Player::loadPlayer(XML_Helper* helper)
{
	TYPE type;
	int t;
	
	helper->getData(t, "type");
	type = static_cast<TYPE>(t);

	switch (type)
	{
		case HUMAN:
			return (new RealPlayer(helper));
		case AI_FAST:
			return (new AI_Fast(helper));
		case AI_DUMMY:
			return (new AI_Dummy(helper));
		case NEUTRAL:
			return (new AI_Dummy(helper));
	}

	return 0;
}

bool Player::load(string tag, XML_Helper* helper)
{
	if (tag != "action")
		return false;
	
	Action* action;
	action = Action::handle_load(helper);
	d_actions.push_back(action);

	return true;
}

// End of file
